@@ -1,4 +1,7 @@
-/* Copyright (c) 1994, Joseph Arceneaux.  All rights reserved
+/*
+   Copyright 1999-2002,2010, Thomas Dickey
+
+   Copyright (c) 1994, Joseph Arceneaux.  All rights reserved
 
    Copyright (c) 1992, Free Software Foundation, Inc.  All rights reserved.
 
@@ -26,27 +29,22 @@
 
 #include "sys.h"
 #include "indent.h"
-#include <ctype.h>
-
-#if defined (HAVE_UNISTD_H)
-#include <unistd.h>
-#endif
-#include <string.h>
-
 
 /* Stuff that needs to be shared with the rest of indent. Documented in
    indent.h.  */
 char *token;
 char *token_end;
+int token_len;
+char *token_buf;
 
 #define alphanum 1
 #define opchar 3
 
 struct templ
-{
-  char *rwd;
-  enum rwcodes rwcode;
-};
+  {
+    const char *rwd;
+    enum rwcodes rwcode;
+  };
 
 /* Pointer to a vector of keywords specified by the user.  */
 static struct templ *user_specials = 0;
@@ -57,144 +55,42 @@ static unsigned int user_specials_max;
 /* Index in user_specials of the first unused entry.  */
 static unsigned int user_specials_idx;
 
-char chartype[128] =
-{				/* this is used to facilitate the decision of
-				   what type (alphanumeric, operator) each
-				   character is */
-  0, 0, 0, 0, 0, 0, 0, 0,
-  0, 0, 0, 0, 0, 0, 0, 0,
-  0, 0, 0, 0, 0, 0, 0, 0,
-  0, 0, 0, 0, 0, 0, 0, 0,
-  0, 3, 0, 0, 1, 3, 3, 0,
-  0, 0, 3, 3, 0, 3, 0, 3,
-  1, 1, 1, 1, 1, 1, 1, 1,
-  1, 1, 0, 0, 3, 3, 3, 3,
-  0, 1, 1, 1, 1, 1, 1, 1,
-  1, 1, 1, 1, 1, 1, 1, 1,
-  1, 1, 1, 1, 1, 1, 1, 1,
-  1, 1, 1, 0, 0, 0, 3, 1,
-  0, 1, 1, 1, 1, 1, 1, 1,
-  1, 1, 1, 1, 1, 1, 1, 1,
-  1, 1, 1, 1, 1, 1, 1, 1,
-  1, 1, 1, 0, 3, 0, 3, 0
-};
+/* this is used to facilitate the decision of what type (alphanumeric,
+ * operator) each character is
+ */
+char chartype[256];
 
-/* C code produced by gperf version 2.5 (GNU C++ version) */
-/* Command-line: gperf -D -c -p -t -T -g -j1 -o -K rwd -N is_reserved indent-c++.gperf  */
+#include "gperf.c"
 
-#define TOTAL_KEYWORDS 39
-#define MIN_WORD_LENGTH 2
-#define MAX_WORD_LENGTH 9
-#define MIN_HASH_VALUE 4
-#define MAX_HASH_VALUE 67
-/* maximum key range = 64, duplicates = 0 */
+/* We do not always increment token_len when we increment buf_ptr, but the
+ * reverse is true.  Make the result a little more readable.
+ */
+#define inc_token_len() token_len += (token_buf == cur_line), ++buf_ptr
 
-#ifdef __GNUC__
-inline
-#endif
-static unsigned int
-hash (char *str, int unsigned len)
+static int
+ident_is_cast (void)
 {
-  static unsigned char asso_values[] =
+  int result;
+  if (parser_state_tos->p_l_follow
+      && !(parser_state_tos->noncast_mask
+	   & 1 << parser_state_tos->p_l_follow))
+    /* inside parens: cast */
     {
-     68, 68, 68, 68, 68, 68, 68, 68, 68, 68,
-     68, 68, 68, 68, 68, 68, 68, 68, 68, 68,
-     68, 68, 68, 68, 68, 68, 68, 68, 68, 68,
-     68, 68, 68, 68, 68, 68, 68, 68, 68, 68,
-     68, 68, 68, 68, 68, 68, 68, 68, 68, 68,
-     68, 68, 68, 68, 68, 68, 68, 68, 68, 68,
-     68, 68, 68, 68, 68, 68, 68, 68, 68, 68,
-     68, 68, 68, 68, 68, 68, 68, 68, 68, 68,
-     68, 68, 68, 68, 68, 68, 68, 68, 68, 68,
-     68, 68, 68, 68, 68, 68, 68, 68, 31,  4,
-      0,  0, 21, 33,  6, 19, 68,  0, 20,  1,
-     23, 30, 14, 68,  6,  7,  0, 10,  7,  5,
-     68, 68, 68, 68, 68, 68, 68, 68,
-    };
-
-  return len + asso_values[(int) str[len - 1]] + asso_values[(int) str[0]];
-}
-
-#ifdef __GNUC__
-inline
-#endif
-struct templ*
-is_reserved (str, len)
-     char *str;
-     unsigned int len;
-{
-  static struct templ wordlist[] =
-    {
-      {"",}, {"",}, {"",}, {"",}, 
-      {"else",  rw_sp_nparen,},
-      {"enum",  rw_struct_like,},
-      {"double",  rw_decl,},
-      {"default",  rw_case,},
-      {"case",  rw_case,},
-      {"const",  rw_decl,},
-      {"while",  rw_sp_paren,},
-      {"void",  rw_decl,},
-      {"short",  rw_decl,},
-      {"struct",  rw_struct_like,},
-      {"char",  rw_decl,},
-      {"volatile",  rw_decl,},
-      {"class",  rw_struct_like,},
-      {"static",  rw_decl,},
-      {"unsigned",  rw_decl,},
-      {"switch",  rw_switch,},
-      {"register",  rw_decl,},
-      {"private",  rw_decl,},
-      {"int",  rw_decl,},
-      {"protected",  rw_decl,},
-      {"public",  rw_decl,},
-      {"inline",  rw_decl,},
-      {"float",  rw_decl,},
-      {"friend",  rw_decl,},
-      {"typedef",  rw_decl,},
-      {"extern",  rw_decl,},
-      {"for",  rw_sp_paren,},
-      {"new",  rw_sizeof,},
-      {"do",  rw_sp_nparen,},
-      {"va_dcl",  rw_decl,},
-      {"sizeof",  rw_sizeof,},
-      {"return",  rw_return,},
-      {"break",  rw_break,},
-      {"",}, 
-      {"union",  rw_struct_like,},
-      {"",}, {"",}, {"",}, 
-      {"if",  rw_sp_paren,},
-      {"",}, 
-      {"operator",  rw_operator,},
-      {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, 
-      {"",}, {"",}, {"",}, 
-      {"long",  rw_decl,},
-      {"",}, 
-      {"global",  rw_decl,},
-      {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, {"",}, 
-      {"goto",  rw_break,},
-    };
-
-  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
-    {
-      int key = hash (str, len);
-
-      if (key <= MAX_HASH_VALUE && key >= 0)
-        {
-          char *s = wordlist[key].rwd;
-
-          if (*s == *str && !strncmp (str + 1, s + 1, len - 1))
-            return &wordlist[key];
-        }
+      parser_state_tos->cast_mask
+	|= 1 << parser_state_tos->p_l_follow;
+      result = true;
     }
-  return 0;
+  else
+    {
+      result = false;
+    }
+  return result;
 }
 
-
-extern int squest;
-
 enum codes
-lexi ()
+lexi (void)
 {
+
   int unary_delim;		/* this is set to 1 if the current token
 				   forces a following operator to be unary */
   static enum codes last_code;	/* the last token type returned */
@@ -211,11 +107,11 @@ lexi ()
   parser_state_tos->col_1 = parser_state_tos->last_nl;
   parser_state_tos->last_nl = false;
 
-  if  (*buf_ptr == ' ' || *buf_ptr == TAB)
+  if (isblank (*buf_ptr))
     {
       parser_state_tos->col_1 = false;
-      while (*buf_ptr == ' ' || *buf_ptr == TAB)
-	if (++buf_ptr >= buf_end)
+      while (isblank (*buf_ptr))
+	if (at_buffer_end (++buf_ptr))
 	  fill_buffer ();
     }
 
@@ -225,26 +121,27 @@ lexi ()
      to a different buffer.  Thus when `token_end' gets set later, it
      may be pointing into a different buffer than `token'. */
   token = buf_ptr;
+  token_len = 0;
+  token_buf = cur_line;
 
   /* Scan an alphanumeric token */
-  if ((! (buf_ptr[0] == 'L' && (buf_ptr[1] == '"' || buf_ptr[1] == '\''))
-          && chartype[(int) *buf_ptr] == alphanum)
-      || (buf_ptr[0] == '.' && isdigit (buf_ptr[1])))
+  if ((!(buf_ptr[0] == 'L' && (buf_ptr[1] == '"' || buf_ptr[1] == '\''))
+       && chartype[UChar (*buf_ptr)] == alphanum)
+      || (buf_ptr[0] == '.' && isdigit (UChar (buf_ptr[1]))))
     {
       /* we have a character or number */
-      struct templ *p;
+      const struct templ *p;
 
-      if (isdigit (*buf_ptr) || (buf_ptr[0] == '.' && isdigit (buf_ptr[1])))
+      if (isdigit (UChar (*buf_ptr))
+	  || (buf_ptr[0] == '.' && isdigit (UChar (buf_ptr[1]))))
 	{
 	  int seendot = 0, seenexp = 0;
 	  if (*buf_ptr == '0' &&
 	      (buf_ptr[1] == 'x' || buf_ptr[1] == 'X'))
 	    {
 	      buf_ptr += 2;
-	      while (isxdigit (*buf_ptr))
-		{
-		  buf_ptr++;
-		}
+	      while (isxdigit (UChar (*buf_ptr)))
+		inc_token_len ();
 	    }
 	  else
 	    while (1)
@@ -256,8 +153,8 @@ lexi ()
 		    else
 		      seendot++;
 		  }
-		buf_ptr++;
-		if (!isdigit (*buf_ptr) && *buf_ptr != '.')
+		inc_token_len ();
+		if (!isdigit (UChar (*buf_ptr)) && *buf_ptr != '.')
 		  {
 		    if ((*buf_ptr != 'E' && *buf_ptr != 'e') || seenexp)
 		      break;
@@ -265,46 +162,46 @@ lexi ()
 		      {
 			seenexp++;
 			seendot++;
-			buf_ptr++;
+			inc_token_len ();
 			if (*buf_ptr == '+' || *buf_ptr == '-')
-			  buf_ptr++;
+			  inc_token_len ();
 		      }
 		  }
 	      }
 
 	  if (*buf_ptr == 'F' || *buf_ptr == 'f'
 	      || *buf_ptr == 'i' || *buf_ptr == 'j')
-	    buf_ptr++;
+	    inc_token_len ();
 	  else
 	    {
 	      if (*buf_ptr == 'U' || *buf_ptr == 'u')
-		buf_ptr++;
+		inc_token_len ();
 	      if (*buf_ptr == 'L' || *buf_ptr == 'l')
-		buf_ptr++;
+		inc_token_len ();
 	      if (*buf_ptr == 'L' || *buf_ptr == 'l')
-		buf_ptr++;
+		inc_token_len ();
 	    }
 	}
       else
-	while (chartype[(int) *buf_ptr] == alphanum)
+	while (chartype[UChar (*buf_ptr)] == alphanum)
 	  {			/* copy it over */
-	    buf_ptr++;
-	    if (buf_ptr >= buf_end)
+	    inc_token_len ();
+	    if (at_buffer_end (buf_ptr))
 	      fill_buffer ();
 	  }
 
       token_end = buf_ptr;
 
-      while (*buf_ptr == ' ' || *buf_ptr == TAB)
+      while (isblank (*buf_ptr))
 	{
-	  if (++buf_ptr >= buf_end)
+	  if (at_buffer_end (++buf_ptr))
 	    fill_buffer ();
 	}
       parser_state_tos->its_a_keyword = false;
       parser_state_tos->sizeof_keyword = false;
 
       /* if last token was 'struct', then this token should be treated
-	 as a declaration */
+         as a declaration */
       if (l_struct)
 	{
 	  l_struct = false;
@@ -320,7 +217,7 @@ lexi ()
       last_code = ident;
 
       /* Check whether the token is a reserved word.  Use perfect hashing... */
-      p = is_reserved (token, token_end - token);
+      p = is_reserved (token, (unsigned) (token_end - token));
 
       if (!p && user_specials != 0)
 	{
@@ -328,8 +225,8 @@ lexi ()
 	       p < &user_specials[0] + user_specials_idx;
 	       p++)
 	    {
-	      char *q = token;
-	      char *r = p->rwd;
+	      const char *q = token;
+	      const char *r = p->rwd;
 
 	      /* This string compare is a little nonstandard because token
 	         ends at the character before token_end and p->rwd is
@@ -361,6 +258,10 @@ lexi ()
 	  enum codes value;
 
 	found_keyword:
+	  if (debug)
+	    {
+	      printf ("keyword \"%.*s\"\n", token_len, token);
+	    }
 	  value = ident;
 	  parser_state_tos->its_a_keyword = true;
 	  parser_state_tos->last_u_d = true;
@@ -380,31 +281,22 @@ lexi ()
 	      break;
 
 	    case rw_struct_like:	/* a "struct" */
-	      if (parser_state_tos->p_l_follow
-		  && !(parser_state_tos->noncast_mask
-		       & 1 << parser_state_tos->p_l_follow))
-		/* inside parens: cast */
+	      if (!ident_is_cast ())
 		{
-		  parser_state_tos->cast_mask
-		    |= 1 << parser_state_tos->p_l_follow;
-		  break;
+		  l_struct = true;
+		  last_code = decl;
+		  value = (decl);
 		}
-	      l_struct = true;
+	      break;
 
 	      /* Next time around, we will want to know that we have had a
 	         'struct' */
 	    case rw_decl:	/* one of the declaration keywords */
-	      if (parser_state_tos->p_l_follow
-		  && !(parser_state_tos->noncast_mask
-		       & 1 << parser_state_tos->p_l_follow))
-		/* inside parens: cast */
+	      if (!ident_is_cast ())
 		{
-		  parser_state_tos->cast_mask
-		    |= 1 << parser_state_tos->p_l_follow;
-		  break;
+		  last_code = decl;
+		  value = (decl);
 		}
-	      last_code = decl;
-	      value = (decl);
 	      break;
 
 	    case rw_sp_paren:	/* if, while, for */
@@ -435,61 +327,61 @@ lexi ()
 	       && parser_state_tos->tos <= 1
 	       && parser_state_tos->ind_level == 0
 	       && parser_state_tos->paren_depth == 0)
-	  {
-	    /* We have found something which might be the name in a function
-	       definition.  */
-	    char *tp;
-	    int paren_count = 1;
+	{
+	  /* We have found something which might be the name in a function
+	     definition.  */
+	  const char *tp;
+	  int paren_count = 1;
 
-	    /* Skip to the matching ')'.  */
-	    for (tp = buf_ptr + 1;
-		 paren_count > 0 && tp < in_prog + in_prog_size;
-		 tp++)
-	      {
-		if (*tp == '(')
-		  paren_count++;
-		if (*tp == ')')
-		  paren_count--;
-		/* Can't occur in parameter list; this way we don't search the
-		   whole file in the case of unbalanced parens.  */
-		if (*tp == ';')
-		  goto not_proc;
-	      }
+	  /* Skip to the matching ')'.  */
+	  for (tp = buf_ptr + 1;
+	       paren_count > 0 && tp < in_prog + in_prog_size;
+	       tp++)
+	    {
+	      if (*tp == '(')
+		paren_count++;
+	      if (*tp == ')')
+		paren_count--;
+	      /* Can't occur in parameter list; this way we don't search the
+	         whole file in the case of unbalanced parens.  */
+	      if (*tp == ';')
+		goto not_proc;
+	    }
 
-	    if (paren_count == 0)
-	      {
-		parser_state_tos->procname = token;
-		parser_state_tos->procname_end = token_end;
+	  if (paren_count == 0)
+	    {
+	      parser_state_tos->procname = token;
+	      parser_state_tos->procname_end = token_end;
 
-		while (isspace (*tp))
-		  tp++;
+	      while (isspace (UChar (*tp)))
+		tp++;
 
-		/* If the next char is ';' or ',' or '(' we have a function
-		   declaration, not a definition.
-		   
-		   I've added '=' to this list to keep from breaking
-		   a non-valid C macro from libc.  -jla */
-		if (*tp != ';' && *tp != ',' && *tp != '(' && *tp != '=')
-		  {
-		    parser_state_tos->in_parameter_declaration = 1;
-		  }
-	      }
+	      /* If the next char is ';' or ',' or '(' we have a function
+	         declaration, not a definition.
 
-	  not_proc:;
-	  }
+	         I've added '=' to this list to keep from breaking
+	         a non-valid C macro from libc.  -jla */
+	      if (*tp != ';' && *tp != ',' && *tp != '(' && *tp != '=')
+		{
+		  parser_state_tos->in_parameter_declaration = 1;
+		}
+	    }
+
+	not_proc:;
+	}
       else if (*buf_ptr == ':' && *(buf_ptr + 1) == ':'
 	       && parser_state_tos->tos <= 1
 	       && parser_state_tos->ind_level == 0
 	       && parser_state_tos->paren_depth == 0)
-	  {
-	    parser_state_tos->classname = token;
-	    parser_state_tos->classname_end = token_end;
-	  }
+	{
+	  parser_state_tos->classname = token;
+	  parser_state_tos->classname_end = token_end;
+	}
       /* The following hack attempts to guess whether or not the
-	 current token is in fact a declaration keyword -- one that
-	 has been typedef'd */
+         current token is in fact a declaration keyword -- one that
+         has been typedef'd */
       else if (((*buf_ptr == '*' && buf_ptr[1] != '=')
-		|| isalpha (*buf_ptr) || *buf_ptr == '_')
+		|| isalpha (UChar (*buf_ptr)) || *buf_ptr == '_')
 	       && !parser_state_tos->p_l_follow
 	       && !parser_state_tos->block_init
 	       && (parser_state_tos->last_token == rparen
@@ -498,15 +390,34 @@ lexi ()
 		   || parser_state_tos->last_token == decl
 		   || parser_state_tos->last_token == lbrace
 		   || parser_state_tos->last_token == start_token))
-	  {
+	{
+	  parser_state_tos->its_a_keyword = true;
+	  parser_state_tos->last_u_d = true;
+	  last_code = decl;
+	  if (parser_state_tos->last_token == cpp_operator)
+	    return overloaded;
+
+	  return decl;
+	}
+      /*
+       * If auto-typedefs, treat any identifier ending with "_t" as a type.
+       */
+      else if (auto_typedefs)
+	{
+	  if (token_len > 2
+	      && !strncmp (token_end - 2, "_t", (size_t) 2))
+	    {
 	      parser_state_tos->its_a_keyword = true;
 	      parser_state_tos->last_u_d = true;
-	      last_code = decl;
-	      if (parser_state_tos->last_token == cpp_operator)
-		  return overloaded;
-
-	      return decl;
-	  }
+	      parser_state_tos->last_rw_depth = parser_state_tos->paren_depth;
+	      if (!ident_is_cast ())
+		{
+		  parser_state_tos->last_rw = rw_decl;
+		  last_code = decl;
+		  return (decl);
+		}
+	    }
+	}
 
       if (last_code == decl)	/* if this is a declared variable, then
 				   following sign is unary */
@@ -527,7 +438,7 @@ lexi ()
      Note that it may be possible for this to kill us--if `fill_buffer'
      at any time switches `buf_ptr' to the other input buffer, `token'
      and `token_end' will point to different storage areas!!! */
-  if (++buf_ptr >= buf_end)
+  if (at_buffer_end (++buf_ptr))
     fill_buffer ();
 
   switch (*token)
@@ -542,7 +453,7 @@ lexi ()
       code = newline;
       break;
 
-    /* Handle wide strings and chars. */
+      /* Handle wide strings and chars. */
     case 'L':
       if (buf_ptr[0] != '"' && buf_ptr[0] != '\'')
 	{
@@ -552,7 +463,7 @@ lexi ()
 	}
 
       qchar = buf_ptr[0];
-      buf_ptr++;
+      inc_token_len ();
       goto handle_string;
 
     case '\'':			/* start of quoted character */
@@ -568,31 +479,29 @@ lexi ()
 	{
 	  if (*buf_ptr == '\\')
 	    {
-	      buf_ptr++;
-	      if (buf_ptr >= buf_end)
+	      inc_token_len ();
+	      if (at_buffer_end (buf_ptr))
 		fill_buffer ();
-	      if (*buf_ptr == EOL)
-		++line_no;
 	      if (*buf_ptr == 0)
 		break;
 	    }
-	  buf_ptr++;
-	  if (buf_ptr >= buf_end)
+	  inc_token_len ();
+	  if (at_buffer_end (buf_ptr))
 	    fill_buffer ();
 	}
 
       if (*buf_ptr == EOL || *buf_ptr == 0)
 	{
-	  WARNING ((qchar == '\''
-		    ? "Unterminated character constant"
-		    : "Unterminated string constant"),
+	  message (1, (qchar == '\''
+		       ? "Unterminated character constant"
+		       : "Unterminated string constant"),
 		   0, 0);
 	}
       else
 	{
 	  /* Advance over end quote char.  */
-	  buf_ptr++;
-	  if (buf_ptr >= buf_end)
+	  inc_token_len ();
+	  if (at_buffer_end (buf_ptr))
 	    fill_buffer ();
 	}
 
@@ -616,13 +525,18 @@ lexi ()
       code = preesc;
 
       /* Make spaces between '#' and the directive be part of
-	 the token if user specified "-lps" */
-      if (leave_preproc_space)
+         the token if user specified "-lps" and did not specify "-ppi" */
+      if (leave_preproc_space && !preprocessor_indentation)
 	{
-	  while (*buf_ptr == ' ' && buf_ptr < buf_end)
-	    buf_ptr++;
+	  while (isblank (*buf_ptr) && !at_buffer_end (buf_ptr))
+	    inc_token_len ();
 	  token_end = buf_ptr;
 	}
+      else
+	{
+	  while (isblank (*buf_ptr) && !at_buffer_end (buf_ptr))
+	    ++buf_ptr;
+	}
       break;
 
     case '?':
@@ -635,14 +549,14 @@ lexi ()
       if (*buf_ptr == ':')
 	{
 	  code = doublecolon;
-	  buf_ptr++;
+	  inc_token_len ();
 	  token_end = buf_ptr;
 	  break;
 	}
 
       code = colon;
       unary_delim = true;
-      if (squest && *e_com != ' ')
+      if (squest && !isblank (*e_com))
 	{
 	  if (e_code == s_code)
 	    parser_state_tos->want_blank = false;
@@ -656,36 +570,29 @@ lexi ()
       code = semicolon;
       break;
 
-    case ('{'):
+    case (L_CURL):
       unary_delim = true;
 
-      /* This check is made in the code for '='.  No one who writes
-         initializers without '=' these days deserves to have indent work on
-         their code (besides which, uncommenting this would screw up anything
-         which assumes that parser_state_tos->block_init really means you are
-         in an initializer.  */
-      /* if (parser_state_tos->in_or_st) parser_state_tos->block_init = 1; */
-
       /* The following neat hack causes the braces in structure
          initializations to be treated as parentheses, thus causing
          initializations to line up correctly, e.g. struct foo bar = {{a, b,
          c}, {1, 2}}; If lparen is returned, token can be used to distinguish
-         between '{' and '(' where necessary.  */
+         between left curly-brace and '(' where necessary.  */
 
       code = parser_state_tos->block_init ? lparen : lbrace;
       break;
 
-    case ('}'):
+    case (R_CURL):
       unary_delim = true;
-      /* The following neat hack is explained under '{' above.  */
+      /* The following neat hack is explained under left curly-brace above.  */
       code = parser_state_tos->block_init ? rparen : rbrace;
 
       break;
 
-    case 014:			/* a form feed */
+    case FORM_FEED:		/* a form feed */
       unary_delim = parser_state_tos->last_u_d;
-      parser_state_tos->last_nl = true;	/* remember this so we can set
-					   'parser_state_tos->col_1' right */
+      parser_state_tos->last_nl = true;		/* remember this so we can set
+						   'parser_state_tos->col_1' right */
       code = form_feed;
       break;
 
@@ -707,7 +614,7 @@ lexi ()
       if (*buf_ptr == token[0])
 	{
 	  /* check for doubled character */
-	  buf_ptr++;
+	  inc_token_len ();
 	  /* buffer overflow will be checked at end of loop */
 	  if (last_code == ident || last_code == rparen)
 	    {
@@ -718,11 +625,11 @@ lexi ()
 	}
       else if (*buf_ptr == '=')
 	/* check for operator += */
-	buf_ptr++;
+	inc_token_len ();
       else if (*buf_ptr == '>')
 	{
 	  /* check for operator -> */
-	  buf_ptr++;
+	  inc_token_len ();
 	  code = struct_delim;
 	}
 
@@ -733,10 +640,10 @@ lexi ()
     case '=':
       if (parser_state_tos->in_or_st &&
 	  parser_state_tos->last_token != cpp_operator)
-	parser_state_tos->block_init = 1;
+	parser_state_tos->block_init = true;
 
       if (*buf_ptr == '=')	/* == */
-	buf_ptr++;
+	inc_token_len ();
       else if (*buf_ptr == '-'
 	       || *buf_ptr == '+'
 	       || *buf_ptr == '*'
@@ -749,13 +656,16 @@ lexi ()
 	     people want to detect and eliminate old style assignments but
 	     they don't want indent to silently change the meaning of their
 	     code).  */
-	  WARNING ("old style assignment ambiguity in \"=%c\".  Assuming \"= %c\"\n",
+	  message (1,
+		   "old style assignment ambiguity in \"=%c\".  Assuming \"= %c\"\n",
 		   (int) *buf_ptr, (int) *buf_ptr);
 	}
 
       code = binary_op;
       unary_delim = true;
       token_end = buf_ptr;
+      if (indent_eqls < 0)
+	indent_eqls = token_col ();
       break;
       /* can drop thru!!! */
 
@@ -767,11 +677,13 @@ lexi ()
          one token, but I don't think that will cause any harm.  */
       while (*buf_ptr == '>' || *buf_ptr == '<' || *buf_ptr == '=')
 	{
-	  if (++buf_ptr >= buf_end)
+	  inc_token_len ();
+	  if (at_buffer_end (buf_ptr))
 	    fill_buffer ();
 	  if (*buf_ptr == '=')
 	    {
-	      if (++buf_ptr >= buf_end)
+	      inc_token_len ();
+	      if (at_buffer_end (buf_ptr))
 		fill_buffer ();
 	    }
 	}
@@ -781,19 +693,6 @@ lexi ()
       token_end = buf_ptr;
       break;
 
-#if 0
-      /* Contributed by Beverly Brown, <beverly@norton.datacube.com> */
-    case '*':
-      /* If the * is part of a declaration such as char *,
-	 we're still a decl */
-      if(parser_state_tos->last_token == decl)
-	{
-	  unary_delim = parser_state_tos->last_u_d;
-	  last_code = decl;
-	  return decl;
-	}
-#endif
-
     default:
       if (token[0] == '/' && (*buf_ptr == '*' || *buf_ptr == '/'))
 	{
@@ -804,7 +703,8 @@ lexi ()
 	  else
 	    code = cplus_comment;
 
-	  if (++buf_ptr >= buf_end)
+	  inc_token_len ();
+	  if (at_buffer_end (buf_ptr))
 	    fill_buffer ();
 
 	  unary_delim = parser_state_tos->last_u_d;
@@ -820,7 +720,8 @@ lexi ()
 	  while (*(buf_ptr - 1) == *buf_ptr || *buf_ptr == '=')
 	    {
 	      /* handle ||, &&, etc, and also things as in int *****i */
-	      if (++buf_ptr >= buf_end)
+	      inc_token_len ();
+	      if (at_buffer_end (buf_ptr))
 		fill_buffer ();
 	    }
 	  code = (parser_state_tos->last_u_d ? unary_op : binary_op);
@@ -837,7 +738,7 @@ lexi ()
       last_code = code;
     }
 
-  if (buf_ptr >= buf_end)
+  if (at_buffer_end (buf_ptr))
     fill_buffer ();
   parser_state_tos->last_u_d = unary_delim;
 
@@ -846,16 +747,54 @@ lexi ()
   return (code);
 }
 
+int
+token_col (void)
+{
+  int col;
+  int n;
+
+  for (n = col = 0; n <= (token - token_buf); ++n)
+    {
+      if (token_buf[n] == '\t')
+	col = (col | 7) + 1;
+      else
+	++col;
+    }
+  if (col == 0)
+    col = 1;
+  return col;
+}
+
+int
+first_token_col (void)
+{
+  int column;
+  int n;
+
+  for (n = column = 0; n <= (buf_ptr - cur_line); ++n)
+    {
+      if (cur_line[n] == TAB)
+	column += tabsize - (column - 1) % tabsize;
+      else if (cur_line[n] == ' ')
+	++column;
+      else
+	break;
+    }
+  if (column == 0)
+    column = 1;
+  return column;
+}
+
 /* Add the given keyword to the keyword table, using val as
    the keyword type */
 
 void
-addkey (char *key, enum rwcodes val)
+addkey (const char *key, enum rwcodes val)
 {
   struct templ *p;
 
   /* Check to see whether key is a reserved word or not. */
-  if (is_reserved (key, strlen (key)) != 0)
+  if (is_reserved (key, (unsigned) strlen (key)) != 0)
     return;
 
   if (user_specials == 0)
@@ -877,3 +816,27 @@ addkey (char *key, enum rwcodes val)
   p->rwcode = val;
   return;
 }
+
+static void
+set_chartype (const char *list, int code)
+{
+  while (*list != 0)
+    {
+      chartype[UChar (*list++)] = (char) code;
+    }
+}
+
+/*
+ * We could build the program source using special tools such as gperf, and
+ * hardcode the character table - or we could generate the tables once at
+ * runtime.  The latter is not much overhead, and is more easily maintained.
+ */
+void
+init_lexi (void)
+{
+  memset (chartype, 0, sizeof (chartype));
+  set_chartype ("$_0123456789", alphanum);
+  set_chartype ("abcdefghijklmnopqrstuvwxyz", alphanum);
+  set_chartype ("ABCDEFGHIJKLMNOPQRSTUVWXYZ", alphanum);
+  set_chartype ("!%&*+-/<=>?^|~", opchar);
+}
